from __future__ import annotations
import concurrent.futures
import json
import logging
import os
import io
import signal
from time import sleep, time
from typing import TYPE_CHECKING

from assemblyline.common.log import init_logging
from assemblyline.common.forge import get_config
from assemblyline.odm.models.config import Config
from assemblyline.remote.datatypes.queues.named import NamedQueue
from assemblyline.remote.datatypes import get_client as get_redis_client

from multiprocessing import Event

if TYPE_CHECKING:
    from redis import Redis


DEL_TIME: int = 60 * 60  # An hour
MAX_QUEUE_LENGTH = 100000
VACUUM_BUFFER_NAME = 'vacuum-file-buffer'

stop_event = Event()


# noinspection PyUnusedLocal
def sigterm_handler(_signum=0, _frame=None):
    stop_event.set()


_last_heartbeat: float = 0


def heartbeat(config: Config):
    global _last_heartbeat
    if _last_heartbeat + 3 < time():
        with io.open(config.logging.heartbeat_file, 'ab'):
            os.utime(config.logging.heartbeat_file)
        _last_heartbeat = time()


logger = logging.getLogger('assemblyline.vacuum')


def main():
    config = get_config()
    signal.signal(signal.SIGTERM, sigterm_handler)

    # Initialize logging
    init_logging('assemblyline.vacuum')
    logger.info('Vacuum starting up...')

    # Initialize cache
    logger.info("Connect to redis...")
    redis = get_redis_client(config.core.redis.nonpersistent.host, config.core.redis.nonpersistent.port, False)
    run(config, redis)


def run(config: Config, redis: Redis):
    vacuum_config = config.core.vacuum

    # connect to workers
    logger.info("Connect to work queue...")
    queue = NamedQueue(VACUUM_BUFFER_NAME, redis)

    logger.info("Load cache...")
    files_list_cache = os.path.join(vacuum_config.list_cache_directory, 'visited.json')
    try:
        with open(files_list_cache, 'r') as handle:
            previous_iteration_files: set[str] = set(json.load(handle))
    except (OSError, json.JSONDecodeError):
        previous_iteration_files = set()

    # Make sure we can access the cache file
    with open(files_list_cache, 'w'):
        pass

    this_iteration_files: list[str] = []
    length = queue.length()

    # Make sure some input is configured
    if not vacuum_config.data_directories:
        logger.error("No input directory configured.")
        return

    logger.info("Starting main loop...")
    while not stop_event.is_set():
        heartbeat(config)
        remove_dir_list = []
        futures: list[concurrent.futures.Future] = []
        with concurrent.futures.ThreadPoolExecutor(20) as pool:
            for data_directory in vacuum_config.data_directories:
                for root, dirs, files in os.walk(data_directory):
                    heartbeat(config)
                    while len(futures) > 50:
                        futures = [f for f in futures if not f.done()]
                        heartbeat(config)
                        sleep(0.1)

                    if length > MAX_QUEUE_LENGTH:
                        while len(futures) > 0:
                            futures = [f for f in futures if not f.done()]
                            heartbeat(config)
                            sleep(0.1)
                        length = queue.length()

                    while length > MAX_QUEUE_LENGTH:
                        logger.warning("Backlog full")
                        length = queue.length()
                        for _ in range(120):
                            heartbeat(config)
                            sleep(1)
                            if stop_event.is_set():
                                break
                        if stop_event.is_set():
                            break

                    if stop_event.is_set():
                        break

                    if not dirs and not files:
                        if root == data_directory:
                            continue

                        cur_time = time()
                        dir_time = os.lstat(root).st_mtime
                        if (cur_time - dir_time) > DEL_TIME:
                            logger.debug('Directory %s marked for removal.' % root)
                            remove_dir_list.append(root)
                        else:
                            logger.debug(f'Directory {root} empty but not old enough. '
                                         f'[{int(cur_time - dir_time)}/{DEL_TIME}]')
                        continue

                    if files:
                        new_file_list = [os.path.join(root, f) for f in files
                                         if not f.startswith(".") and not f.endswith('.bad')]
                        new_files = set(new_file_list) - previous_iteration_files

                        if new_files:
                            futures.append(pool.submit(queue.push, *new_files))
                            # queue.push(*new_files)
                            length += len(new_files)
                            this_iteration_files.extend(new_files)

        with open(files_list_cache, 'w') as handle:
            json.dump(this_iteration_files, handle)

        previous_iteration_files = set(this_iteration_files)
        this_iteration_files = []

        for d in remove_dir_list:
            logger.debug("Removing empty directory: %s" % d)
            # noinspection PyBroadException
            try:
                os.rmdir(d)
            except Exception:
                pass

        if not stop_event.is_set():
            sleep(5)

    logger.info('Good bye!')


if __name__ == '__main__':
    main()



#!/usr/bin/env python
import elasticapm
import os
import tempfile

from assemblyline.common import forge
from assemblyline.common.archiving import ARCHIVE_QUEUE_NAME
from assemblyline.common.metrics import MetricsFactory
from assemblyline.datastore.collection import ESCollection, Index
from assemblyline.odm.messages.archive_heartbeat import Metrics
from assemblyline.odm.models.submission import Submission
from assemblyline.remote.datatypes import get_client
from assemblyline.remote.datatypes.queues.named import NamedQueue

from assemblyline_core.server_base import ServerBase


class SubmissionNotFound(Exception):
    pass


class Archiver(ServerBase):
    def __init__(self):
        super().__init__('assemblyline.archiver')
        self.apm_client = None
        self.counter = None

        if self.config.datastore.archive.enabled:
            # Publish counters to the metrics sink.
            self.counter = MetricsFactory('archiver', Metrics)
            self.datastore = forge.get_datastore(self.config, archive_access=True)
            self.filestore = forge.get_filestore(config=self.config)
            self.archivestore = forge.get_archivestore(config=self.config)
            self.persistent_redis = get_client(
                host=self.config.core.redis.persistent.host,
                port=self.config.core.redis.persistent.port,
                private=False,
            )

            self.archive_queue: NamedQueue[dict] = NamedQueue(ARCHIVE_QUEUE_NAME, self.persistent_redis)
            if self.config.core.metrics.apm_server.server_url is not None:
                self.log.info(f"Exporting application metrics to: {self.config.core.metrics.apm_server.server_url}")
                elasticapm.instrument()
                self.apm_client = forge.get_apm_client("archiver")
        else:
            self.log.warning("Archive is not enabled in the config, no need to run archiver.")
            exit()

    def stop(self):
        if self.counter:
            self.counter.stop()

        if self.apm_client:
            elasticapm.uninstrument()
        super().stop()

    def run_once(self):
        message = self.archive_queue.pop(timeout=1)

        # If there is no alert bail out
        if not message:
            return
        else:
            try:
                archive_type, type_id, delete_after = message
                self.counter.increment('received')
            except Exception:
                self.log.error(f"Invalid message received: {message}")
                return

        # Start of process alert transaction
        if self.apm_client:
            self.apm_client.begin_transaction('Process archive message')

        try:
            if archive_type == "submission":
                self.counter.increment('submission')
                # Load submission
                submission: Submission = self.datastore.submission.get_if_exists(type_id)
                if not submission:
                    raise SubmissionNotFound(type_id)

                self.datastore.submission.archive(type_id, delete_after=delete_after)
                if not delete_after:
                    self.datastore.submission.update(type_id, [(ESCollection.UPDATE_SET, 'archived', True)],
                                                     index_type=Index.HOT)

                # Gather list of files and archives them
                files = {f.sha256 for f in submission.files}
                files.update(self.datastore.get_file_list_from_keys(submission.results, supplementary=True))
                for sha256 in files:
                    self.counter.increment('file')
                    self.datastore.file.archive(sha256, delete_after=delete_after, allow_missing=True)
                    if self.filestore != self.archivestore:
                        with tempfile.NamedTemporaryFile() as buf:
                            self.filestore.download(sha256, buf.name)
                            try:
                                if os.path.getsize(buf.name):
                                    self.archivestore.upload(buf.name, sha256)
                            except Exception as e:
                                self.log.error(
                                    f"Could not copy file {sha256} from the filestore to the archivestore. ({e})")

                # Archive associated results (Skip emptys)
                for r in submission.results:
                    if not r.endswith(".e"):
                        self.counter.increment('result')
                        self.datastore.result.archive(r, delete_after=delete_after, allow_missing=True)

                # End of process alert transaction (success)
                self.log.info(f"Successfully archived submission '{type_id}'.")
                if self.apm_client:
                    self.apm_client.end_transaction(archive_type, 'success')

            # Invalid archiving type
            else:
                self.counter.increment('invalid')
                self.log.warning(f"'{archive_type}' is not a valid archive type.")
                # End of process alert transaction (success)
                if self.apm_client:
                    self.apm_client.end_transaction(archive_type, 'invalid')

        except SubmissionNotFound:
            self.counter.increment('not_found')
            self.log.warning(f"Could not archive {archive_type} '{type_id}'. It was not found in the system.")
            # End of process alert transaction (failure)
            if self.apm_client:
                self.apm_client.end_transaction(archive_type, 'not_found')

        except Exception:  # pylint: disable=W0703
            self.counter.increment('exception')
            self.log.exception(f'Unhandled exception processing {archive_type} ID: {type_id}')

            # End of process alert transaction (failure)
            if self.apm_client:
                self.apm_client.end_transaction(archive_type, 'exception')

    def try_run(self):
        while self.running:
            self.heartbeat()
            self.run_once()


if __name__ == "__main__":
    with Archiver() as archiver:
        archiver.serve_forever()
